local VERSION = tonumber(("$Revision: 53497 $"):match("%d+"))

Spyglass = AceLibrary("AceAddon-2.0"):new("AceDB-2.0", "AceConsole-2.0",
                                          "AceEvent-2.0", "AceHook-2.1")
local Spyglass, self = Spyglass, Spyglass

Spyglass.version = "r"..VERSION
Spyglass.revision = VERSION

Spyglass:RegisterDB('SpyglassDB')
Spyglass:RegisterDefaults('profile', {
    showBonuses = true,
    lastInspected = nil,
    storeEquipment = false,
    storedEquipment = {}
})

local L = AceLibrary("AceLocale-2.2"):new("Spyglass")
local Tablet = AceLibrary("Tablet-2.0")

L:RegisterTranslations("enUS", function() return {
	["Addon for looking deeper into an inspected person's gear"] = true,
	['Show Inspection Window'] = true,
	['Show the inspection window. (displaying the last person inspected)'] = true,
	['Store Data'] = true,
	['Keep user data.'] = true,
	['Enabled'] = true,
	['Disabled'] = true,
	['Bonus Summary'] = true,
	['Fonts\\FRIZQT__.TTF'] = true,
	['Last Inspected Unit'] = true,
	['nobody'] = true,
	['Saved Units'] = true,
	["|cffeda55fClick|r to toggle the inspect window."] = true,
	['Attributes'] = true,
	['Resistance'] = true,
	['Skills'] = true,
	['Combat'] = true,
	['Defenses'] = true,
	['Spells'] = true,
	['Health and Mana'] = true,
	["%d hp5"] = true,
	["%d mp5"] = true,
	['bonuses not passed?'] = true,
} end)

L:RegisterTranslations("koKR", function() return {
	["Addon for looking deeper into an inspected person's gear"] = "살펴보기 기능에 대한 추가 정보 재공 애드온",
	['Show Inspection Window'] = '살펴보기 창 표시',
	['Show the inspection window. (displaying the last person inspected)'] = '살펴보기 창을 표시합니다. (지난 살펴보기 대상을 표시합니다)',
	['Store Data'] = '저장',
	['Keep user data.'] = '살펴본 대상의 정보를 저장합니다.',
	['Enabled'] = '사용',
	['Disabled'] = '사용 안함',
	['Bonus Summary'] = '추가 능력치 정보',
	['Fonts\\FRIZQT__.TTF'] = 'Fonts\\2002.TTF',
	['Last Inspected Unit'] = '지난 살펴보기 대상',
	['nobody'] = '없음',
	['Saved Units'] = '저장된 대상',
	["|cffeda55fClick|r to toggle the inspect window."] = "|cffeda55f클릭|r시 살펴보기 창을 토글합니다.",
	['Attributes'] = '능력치',
	['Resistance'] = '저항',
	['Skills'] = '숙련도',
	['Combat'] = '전투척도',
	['Defenses'] = '방어도',
	['Spells'] = '주문',
	['Health and Mana'] = '생명력과 마나',
	["%d hp5"] = '5초당 %d의 생명력',
	["%d mp5"] = '5초당 %d의 마나',
--	['bonuses not passed?'] = true,
} end)


-- TODO: find something that provides these -for- me
local slots = {
	["Head"] = true,
	["Neck"] = true,
	["Shoulder"] = true,
	["Shirt"] = true,
	["Chest"] = true,
	["Waist"] = true,
	["Legs"] = true,
	["Feet"] = true,
	["Wrist"] = true,
	["Hands"] = true,
	["Finger0"] = true,
	["Finger1"] = true,
	["Trinket0"] = true,
	["Trinket1"] = true,
	["Back"] = true,
	["MainHand"] = true,
	["SecondaryHand"] = true,
	["Ranged"] = true,
	["Tabard"] = true,
}

Spyglass.options = {
    type = 'group',
    icon = [[Interface\Icons\INV_Misc_Spyglass_03]],
    desc = L["Addon for looking deeper into an inspected person's gear"],
    args = {
        show = {
            type = 'execute',
            name = L['Show Inspection Window'],
            desc = L['Show the inspection window. (displaying the last person inspected)'],
            func = function()
                Spyglass:ShowWindow()
            end,
            order = 1,
        },
        
        spacer = {
            type = 'header',
            name = ' ',
            order = 2,
        },
        
        keep = {
            type = 'toggle',
            name = L['Store Data'],
            desc = L['Keep user data.'],
            get = function() return Spyglass.db.profile.storeEquipment end,
            set = 'SetSavedEquipment',
            map = { [true]=L['Enabled'], [false]=L['Disabled'] },
            order = 3,
        },
    },
}

function Spyglass:OnInitialize()
	for s in pairs(slots) do
		slots[s] = GetInventorySlotInfo (s.."Slot")
	end
	
	self.unit = 'player'
	self.unitName = UnitName('player')
	self.currentInfo = {}
	self.inInspectRange = true
	
	self.frame = nil
	self.model = nil
	self.doll  = nil
	
	-- Set the binding header name
	_G['BINDING_HEADER_Spyglass'] = 'Spyglass'
	
	if self.db.profile.storeEquipment then
	    self.unitCache = self.db.profile.storedEquipment
	else
	    self.unitCache = {}
	end
	
	-- metatable?
end

function Spyglass:ADDON_LOADED(addonName)
    if addonName == 'Blizzard_InspectUI' then
        --self:Print("Blizzard's inspect ui has loaded!")
        --self:Print('Taking a closer look...')
        self:UnregisterEvent('ADDON_LOADED')
        
        -- Local references to the important frames
        self.frame = _G['InspectFrame']
        self.model = _G['InspectModelFrame']
        self.doll = _G['InspectPaperDollFrame']
        
        self:LatchOn()
    end
end

function Spyglass:OnEnable(first)
    if not self.frame then
        self:RegisterEvent('ADDON_LOADED')
    else
        self:LatchOn()
    end
    
	self:RegisterChatCommand('/spyglass', '/spy', self.options)
end

function Spyglass:OnDisable()
    
end

-- Added frames
local infoText = nil
local moveFrame = nil
local summaryFrame = nil
local summaryTextL = nil
local summaryTextR = nil

function Spyglass:GlueMovementFrame()
    local moveFrame = self.movement
    local playerModel = self.model
    
    if not moveFrame then
        moveFrame = CreateFrame("Frame", "InspectDragger", playerModel)
        moveFrame:SetFrameLevel(self.frame:GetFrameLevel() + 1)
        
        -- Scales in OnMouseUp multiplied by scales in OnUpdate == 1.
        -- This is to ensure 1/1 pixel/movement
        
        moveFrame:SetScript("OnMouseDown", function()
            if arg1 == 'LeftButton' then
                this:StartMoving()
                this.isRotating = true
            elseif arg1 == 'RightButton' then
                this:StartMoving()
                this.isMoving = true
            end
        end)
        
        moveFrame:SetScript("OnMouseUp", function()
            local p = playerModel
            if arg1 == 'LeftButton' then
                this.isRotating = false
                this.rotation = p.rotation * 50
            elseif arg1 == 'RightButton' then
                this.isMoving = false
                this.pos_x = p.pos_x * 100
                this.pos_y = p.pos_y * 100
            else
                return
            end
            this:StopMovingOrSizing()
            this:SetAllPoints()
        end)
        
        moveFrame:SetScript("OnUpdate", function()
            local p = playerModel
            if this.isRotating then
                p.rotation = (this:GetCenter() + this.rotation
                              - p:GetCenter()) * 0.02
                p:SetRotation(p.rotation)
            end
            if this.isMoving then
                p.pos_x = (this:GetCenter() + this.pos_x - p:GetCenter()) * 0.01
                p.pos_y = (this:GetTop() + this.pos_y - p:GetTop()) * 0.01
                p:SetPosition(p.zoom, p.pos_x, p.pos_y)
            end
        end)
        
        moveFrame:SetScript("OnMouseWheel", function()
            local p = playerModel
            if arg1 > 0 then
                p.zoom = p.zoom + 0.5
            else
                p.zoom = p.zoom - 0.5
            end
            p:SetPosition(p.zoom, p.pos_x, p.pos_y)
        end)
        
        moveFrame:SetScript("OnShow", function()
            this:SetAllPoints()
        end)
        
        self.movement = moveFrame
    end
    
    moveFrame:EnableMouse(true)
    moveFrame:EnableMouseWheel(true)
    moveFrame:SetMovable(true)
    
    playerModel.rotation = 0.61 -- FROM: InspectModelFrame_OnLoad
    playerModel.pos_x = 0
    playerModel.pos_y = 0
    playerModel.zoom = 0
    
    moveFrame.pan_x = 0
    moveFrame.pan_y = 0
    moveFrame.rotation = playerModel.rotation * 50 -- to get the right rotation
    
    playerModel:SetRotation(playerModel.rotation)
    playerModel:RefreshUnit()
end

local old_onUpdate = nil

function Spyglass:LatchOn()
    -- Attach player rank
    if not self.infoText then
        self.infoText = self.doll:CreateFontString('InspectInfoText',
                                                   'BACKGROUND',
                                                   'GameFontNormalSmall')
        self.infoText:SetPoint('TOP', 'InspectLevelText', 'BOTTOM', 0, -1)
    end
    
    self:GlueMovementFrame()
    
    -- Glue the Summary Frame
    self:GlueSummaryFrame()
    
    -- Glue Tooltip Handling
    self:SecureHook('InspectPaperDollItemSlotButton_OnEnter', 'ButtonOnEnter')
    
    -- Unregister certain events the original used, but we don't want it to use
    self.frame:UnregisterEvent('PLAYER_TARGET_CHANGED')
    self.frame:UnregisterEvent('PARTY_MEMBERS_CHANGED')
    
    for slotName in pairs(slots) do
        _G['Inspect'..slotName..'Slot']:UnregisterEvent('UNIT_INVENTORY_CHANGED')
    end
    
    -- Shifting the event handling over to us
    self:RegisterEvent('PLAYER_TARGET_CHANGED', 'TargetChanged')
    self:RegisterEvent('PARTY_MEMBERS_CHANGED', 'TargetChanged')
    
    self:RegisterEvent('UNIT_INVENTORY_CHANGED', 'UnitChanged')
    self:RegisterEvent('UNIT_NAME_UPDATE', 'UnitChanged')
    
    -- Hooks for adding extra information to the frames
    self:SecureHook('InspectFrame_UnitChanged', 'UnitChanged')
    self:SecureHook('InspectFrame_OnShow', 'UnitChanged')
    self:SecureHook('InspectFrame_Show', 'SetInspectUnit')
    
    -- Prevent closing the InspectFrame when we go out of inspect range
    --   Sloppy, but HookSecure can't overwrite, and we need to.
    if not self.oldOnUpdate then
        self.oldOnUpdate = _G['InspectFrame_OnUpdate']
        _G['InspectFrame_OnUpdate'] = function() end
    end
    
    _G['InspectModelRotateLeftButton']:Hide()
    _G['InspectModelRotateRightButton']:Hide()
    
    --self:Print('Latched on.')
end


function Spyglass:LetGo()
    --self:Print("I'll never let go! ...woops.")
    if self.oldOnUpdate then
        _G['InspectFrame_OnUpdate'] = self.oldOnUpdate
        self.oldOnUpdate = nil
    end
end

function Spyglass:SetSavedEquipment(enable)
    local prof = self.db.profile
    if enable then
        prof.storedEquipment = self.unitCache
    else
        self.unitCache = prof.storedEquipment
        prof.storedEquipment = {}
    end
    prof.storeEquipment = enable

end

local function LoadInspectInfo(unit, info)
    local association, title = GetGuildInfo(unit)
    
    -- try to re-use old info tables
    if not info then
        info = {
            name = UnitName(unit),
            level = UnitLevel(unit),
            race = UnitRace(unit),
            class = UnitClass(unit),
            association = association,
            title = title,
            equipment = {},
        }
    else
        -- only update: level, association and title
        info.level = UnitLevel(unit)
        info.association = association
        info.title = title
    end
    
    -- Until IBL's GetUnitEquipment doesn't hide InspectFrame
    local equipment = info.equipment
    for slot, id in pairs(slots) do
        equipment[slot] = GetInventoryItemLink(unit, id)
    end
    
    return info
end

function Spyglass:GetUnitInfo(unitName, update)
    local info = self.unitCache[unitName]
    if not update and info then return info end
    
    info = LoadInspectInfo(self.frame.unit, info)
    self.unitCache[unitName] = info
    return info
end

function Spyglass:GetCachedPlayerNames()
    -- Create and sort a list of unit names
    local sortedNames = {}
    for name in pairs(self.unitCache) do
        sortedNames[#sortedNames+1] = name
    end
    table.sort(sortedNames)
    return sortedNames
end

local function CheckCache(unitName, backup)
    -- returns: index, if the given names were in cache
    if Spyglass.unitCache[unitName] then
        return unitName, true
    elseif Spyglass.unitCache[backup] then
        return backup, true
    end
    
    return UnitName('player'), false
end

function Spyglass:FinalizeSetInspect(unit, unitName)
    --self:Print('Setting unit to:', unit,', with name:', unitName)
    
    if not self.inInspectRange then
        local name, cached = CheckCache(unitName, self.unitName)
        if not cached then self.inInspectRange = true end
        self.frame.unit = 'player'
        self.unitName = name
    else
        self.frame.unit = unit
        self.unitName = unitName
    end
    
    self.currentInfo = self:GetUnitInfo(self.unitName, self.inInspectRange)
    --self:Print('REALLY Setting unit to:', self.frame.unit,', with name:', self.unitName)
    
    if not self.frame:IsShown() then
        self.frame:Show()
    else
        -- A little messy, but we need to make InspectUI think it was called
        --   from itself.
        --this = self.frame
        -- The above breaks if we were called from the drop-down menus
        this.unit = self.frame.unit
        securecall('InspectFrame_UnitChanged')
    end
end

function Spyglass:SetInspectPlayer(unitName)
    if not self.frame then InspectFrame_LoadUI() end
    
    self.inInspectRange =
        (UnitName('target') == unitName and CanInspect('target'))
    
    self:FinalizeSetInspect('target', unitName)
end

function Spyglass:SetInspectUnit(unit)
    -- If we switched to a target that we can't inspect, including targeting
    --   nobody at all, make InspectUI think we're inspecting outself.
    --   'player' was chosen because it's guaranteed to work with
    --   InspectFrame 100% of the time. We also mark an internal flag that
    --   we should use our own info
    if not self.frame then InspectFrame_LoadUI() end
    
    unit = unit or 'target'
    self.inInspectRange = CanInspect(unit)
    
    self:FinalizeSetInspect(unit, UnitName(unit))
end

function Spyglass:ShowWindow()
    -- Show the inspection window if it's not shown already; otherwise, hide it
    if self.frame and self.frame:IsShown() then
        HideUIPanel(self.frame)
    else
        self:SetInspectUnit('target')
    end
end


-- Player's target changed
function Spyglass:TargetChanged()
    if not self.frame:IsShown() then return end
    
    self:SetInspectUnit('target')
end

-- Unit changed, so set per-unit information
function Spyglass:UnitChanged()
    if not self.frame:IsShown() then return end
    
    local info = self.currentInfo
    
    -- Update Text
    _G['InspectNameText']:SetText(self.unitName)
    _G['InspectLevelText']:SetText(format(PLAYER_LEVEL, info.level, info.race,
                                          info.class))
    
    if info.association and info.title then
        self.infoText:SetText(format(GUILD_TITLE_TEMPLATE, info.title,
                                     '<'..info.association..'>'))
    else
        self.infoText:SetText('')
    end
    
    -- Update Model
    if not self.inInspectRange then
        self.model:Hide()
    else
        self.model:Show()
    end
    
    -- Update Items
    self:UpdateButtons(info, slots)
    
    -- Show Bonuses
    if self.db.profile.showBonuses then
        self:Summarize(info)
    end
    
    self.db.profile.lastInspected = self.unitName
end

Spyglass_FuBar = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0",
                                                "AceConsole-2.0",
                                                "FuBarPlugin-2.0")
local Spyglass_FuBar = Spyglass_FuBar

Spyglass_FuBar.name = 'Spyglass'
Spyglass_FuBar.hasIcon = [[Interface\Icons\INV_Misc_Spyglass_03]]
Spyglass_FuBar.hasNoColor = true
Spyglass_FuBar.defaultMinimapPosition = 285
Spyglass_FuBar.clickableTooltip = true

Spyglass_FuBar.OnMenuRequest = Spyglass.options


function Spyglass_FuBar:OnTooltipClick(name)
    Spyglass:SetInspectPlayer(name)
end

function Spyglass_FuBar:OnTooltipUpdate()
    local cat = Tablet:AddCategory(
        'columns', 2,
        'text', L['Last Inspected Unit']
    )
    
    cat:AddLine(
        'text2', Spyglass.db.profile.lastInspected or L['nobody'],
        'func', Spyglass.SetInspectPlayer,
        'arg1', Spyglass,
        'arg2', Spyglass.db.profile.lastInspected
    )
    
    cat = Tablet:AddCategory(
        'columns', 2,
        'text', L['Saved Units'],
        'child_func', Spyglass.SetInspectPlayer,
        'child_arg1', Spyglass
    )
    
    local info
    for i, name in ipairs(Spyglass:GetCachedPlayerNames()) do
        info = Spyglass:GetUnitInfo(name)
        cat:AddLine('text', name, 'text2', info.level..' '..info.class, 'arg2', name)
    end
    
    Tablet:SetHint(L["|cffeda55fClick|r to toggle the inspect window."])
end

function Spyglass_FuBar:OnClick(button)
    Spyglass:ShowWindow()
end

